//
//  TGraphicSearch.cs
//  Copyright 2005 Takeshi Yoneki / OSTRA / OSTRACISM CO., All rights reserved.
//
using System;
using System.IO;
using System.Collections.Specialized;
using System.Threading;
using System.Text;

namespace GGG4W {
	
	/// <summary>
	/// TGraphicSearch  摜NX
	/// </summary>
	public class TGraphicSearch
	{
		public delegate void HitItemCallbackHandler(TFileInfo info);
		public delegate void ProgressCallbackHandler(TFileInfo info);
		public delegate void SearchEndCallbackHandler();

		public TFileInfo m_target = new TFileInfo("");
		public string m_topDirPath = "";
		public string m_indexFilePath = "";
		public HitItemCallbackHandler HitItemCallback = null;
		public ProgressCallbackHandler ProgressCallback = null;
		public SearchEndCallbackHandler SearchEndCallback = null;
		public int m_bits = 3;
		public int m_matchRate = 70;
		public int m_pixelNumber = 8;

		public const int SEARCH_DIRECTORY = 0;
		public const int SEARCH_INDEXFILE = 1;
		public int m_searchMethod = SEARCH_DIRECTORY;

		Thread m_work = null;
		bool m_continue = false;

		public TGraphicSearch()
		{
			// 
			// TODO: RXgN^ WbNɒǉĂB
			//
		}
		
		/// <summary>
		/// ݒt@CT|[g
		/// </summary>
		/// <param name="setting"></param>
		public void LoadSetting(ref TSetting setting) {
			m_target.m_filePath = setting.targetFilePath;
			m_topDirPath = setting.topDirPath;
			m_indexFilePath = setting.indexFilePath;
			m_bits = setting.bits;
			m_matchRate = setting.matchRate;
			m_pixelNumber = setting.pixelNumber;
			m_searchMethod = setting.searchMethod;
		}

		public void SaveSetting(ref TSetting setting) {
			setting.targetFilePath = m_target.m_filePath;
			setting.topDirPath = m_topDirPath;
			setting.indexFilePath = m_indexFilePath;
			setting.bits = m_bits;
			setting.matchRate = m_matchRate;
			//setting.pixelNumber = m_pixelNumber;
			setting.searchMethod = m_searchMethod;
		}

		/// <summary>
		/// Jn
		/// </summary>
		/// <returns></returns>
		public bool BeginSearch() {
			bool good = true;
			if (m_work != null)
				good = false;
			if (good)
				if (m_target == null)
					good = false;
			if (good)
				if (m_topDirPath == null && m_indexFilePath == null)
					good = false;
			if (good)
				if (m_searchMethod == SEARCH_INDEXFILE)
					good = CheckIndexFile(m_indexFilePath);
			if (good)
				if (m_target.MakeThumbnail(m_pixelNumber) == false)
					good = false;

			if (good == false) {
				if (SearchEndCallback != null)
					SearchEndCallback();
				return false;
			}

			m_continue = true;
			if (m_searchMethod == SEARCH_DIRECTORY)
				m_work = new Thread(new ThreadStart(SearchDirectoryThread));
			else if (m_searchMethod == SEARCH_INDEXFILE)
				m_work = new Thread(new ThreadStart(SearchIndexFileThread));
			m_work.Start();
			return true;
		}

		/// <summary>
		/// ƒ
		/// </summary>
		/// <returns></returns>
		public bool IsWorking() {
			return m_work != null;
		}

		/// <summary>
		/// ~v
		/// </summary>
		public void Stop() {
			lock(this) {
				m_continue = false;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public void Join() {
			Stop();
			while (IsWorking()) {
				Thread.Sleep(0);
			}
		}

		/// <summary>
		/// fBNgXbh
		/// </summary>
		void SearchDirectoryThread() {
			m_target.MakeThumbnail(m_pixelNumber);
			SearchDirectorySub(m_topDirPath);
			m_work = null;
			if (SearchEndCallback != null)
				SearchEndCallback();
		}

		bool SearchDirectorySub(string dirPath) {
			string[] files = Directory.GetFiles(dirPath);
			foreach (string path in files) {
				if (m_continue == false)
					return false;
				TFileInfo current = new TFileInfo(path);
				if (ProgressCallback != null)
					ProgressCallback(current);
				if (current.MakeThumbnail(m_pixelNumber) == false)
					continue;
				if (m_target.IsEqualThumbnail(m_bits, m_matchRate, ref current)) {
					if (HitItemCallback != null)
						HitItemCallback(current);
				}
			}
			string[] dirs = Directory.GetDirectories(dirPath);
			foreach (string path in dirs) {
				if (SearchDirectorySub(path) == false)
					return false;
			}
			return true;
		}

		/// <summary>
		/// CfbNXt@CXbh
		/// </summary>
		void SearchIndexFileThread() {
			FileStream fs = new FileStream(m_indexFilePath, FileMode.Open);
			BinaryReader br = new BinaryReader(fs);

			TChunk ck = new TChunk();
			ck.Read(ref br);
			TChunkHead ckhd = new TChunkHead(ref ck);
			m_pixelNumber = ckhd.m_pixelNumber;
			m_target.MakeThumbnail(m_pixelNumber);

			string filePath = "";
			for (;;) {
				if (fs.Length - fs.Position < 8) {
					break;
				}
				if (m_continue == false)
					break;
				ck = new TChunk();
				ck.Read(ref br);
				if (ck.m_ckID == TChunk.CKID_PATH) {
					TChunkPath ckpt = new TChunkPath(ref ck);
					filePath = ckpt.m_path;
				}
				else if (ck.m_ckID == TChunk.CKID_THUMB) {
					TChunkThumb cktb = new TChunkThumb(ref ck);
					if (filePath != null && filePath.Length > 0) {
						TFileInfo current = new TFileInfo();
						current.m_filePath = filePath;
						current.m_thumbnail = cktb.m_thumbnail;
						current.m_thumbnailCount = m_pixelNumber * m_pixelNumber;
						if (m_target.IsEqualThumbnail(m_bits, m_matchRate, ref current)) {
							if (HitItemCallback != null)
								HitItemCallback(current);
						}
					}
				}
				else {
				}
			}

			br.Close();
			fs.Close();

			m_work = null;
			if (SearchEndCallback != null)
				SearchEndCallback();
		}

		/// <summary>
		/// CfbNXt@C`FbN
		/// </summary>
		/// <param name="path"></param>
		/// <returns></returns>
		public bool CheckIndexFile(string path) {
			bool result = false;
			FileStream fs = null;
			BinaryReader br = null;
			try {
				fs = new FileStream(path, FileMode.Open);
				br = new BinaryReader(fs);
				byte[] buff = new byte[4];
				br.Read(buff, 0 ,4);
				UTF8Encoding uni = new UTF8Encoding();
				string ckID = uni.GetString(buff);
				if (ckID == "GGHD") {
					fs.Position = 0;
					TChunk ck = new TChunk();
					ck.Read(ref br);
					TChunkHead ckhd = new TChunkHead(ref ck);
					m_pixelNumber = ckhd.m_pixelNumber;
					result = true;
				}
			}
			catch {
			}
			if (br != null)
				br.Close();
			if (fs != null)
				fs.Close();
			return result;
		}

		/// <summary>
		/// CfbNXt@Cۑ
		/// </summary>
		/// <returns></returns>
		public bool MakeIndexFile() {
			bool good = true;
			if (m_work != null)
				good = false;
			if (good)
				if (m_topDirPath == null || m_topDirPath.Length == 0)
					good = false;

			if (good == false) {
				if (SearchEndCallback != null)
					SearchEndCallback();
				return false;
			}

			m_continue = true;
			m_work = new Thread(new ThreadStart(MakeIndexFileThread));
			m_work.Start();

			return true;
		}

		/// <summary>
		/// CfbNXt@CۑXbh
		/// </summary>
		void MakeIndexFileThread() {
			FileStream fs = new FileStream(m_indexFilePath, FileMode.Create);
			BinaryWriter bw = new BinaryWriter(fs);
			TChunkHead ckhd = new TChunkHead();
			ckhd.m_pixelNumber = m_pixelNumber;
			ckhd.Write(ref bw);
			MakeIndexFileSub(m_topDirPath, ref bw);
			bw.Close();
			fs.Close();
			m_work = null;
			if (SearchEndCallback != null)
				SearchEndCallback();
		}

		bool MakeIndexFileSub(string dirPath, ref BinaryWriter bw) {
			string[] files = Directory.GetFiles(dirPath);
			foreach (string path in files) {
				if (m_continue == false)
					return false;
				if (string.Compare(path, m_indexFilePath, true) == 0)
					continue;
				TFileInfo current = new TFileInfo(path);
				if (ProgressCallback != null)
					ProgressCallback(current);
				if (current.MakeThumbnail(m_pixelNumber) == false) {
					continue;
				}

				// 
				TChunkPath ckpt = new TChunkPath();
				ckpt.m_path = current.m_filePath;
				ckpt.Write(ref bw);
				TChunkThumb cktb = new TChunkThumb();
				cktb.m_thumbnail = current.m_thumbnail;
				cktb.Write(ref bw);

			}
			string[] dirs = Directory.GetDirectories(dirPath);
			foreach (string path in dirs) {
				if (MakeIndexFileSub(path, ref bw) == false)
					return false;
			}
			return true;			
		}

	}
}
